/*************************************************************************\
* Copyright (c) 2008 The University of Chicago LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*
 * APS Event Receiver Time Provider
 * APS Event Broadcast Time Provdier
 */
#include <epicsStdioRedirect.h>
#include <epicsGeneralTime.h>
#include <generalTimeSup.h>
#include <epicsStdlib.h>
#include <cantProceed.h>
#include <string.h>
#include <math.h>
#include <iocsh.h>

#include <time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>

#include <epicsInterrupt.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <epicsExport.h>
#include <errlog.h>
#include <osiSock.h>
#include "devApsEr.h"
#include "drvTS.h"

/*
 * Assume that broadcaster is dead if nothing received in this time
 */
#define BROADCAST_TIMEOUT_SECONDS   120

/*
 * Event system parameters
 */
#define EVENT_SYSTEM_CLOCK_RATE          1000
#define ER_RESET_TICKS_EVENT             0x7d
#define NANOSECONDS_PER_EVENT_CLOCK_TICK (1000000000/EVENT_SYSTEM_CLOCK_RATE)

/*
 * Broadcasts from event master
 */
#define TS_MAGIC ('T'<<24|'S'<<16|'d'<<8|'r')
#define TS_SLAVE_PORT 18322
#define TS_TIME_OUT_MS 250
#define TS_STAMP_FROM_UNIX 0x0001
typedef enum { TS_time_request, TS_sync_request, TS_sync_msg } TStype;
struct TSstampTransStruct {
    unsigned long magic;           /* identifier */
    TStype type;                   /* transaction type */
    struct timespec master_time;   /* master time stamp - last sync time */
    struct timespec current_time;  /* master current time stamp 1990 epoch */
    struct timespec unix_time;     /* time using 1970 epoch */
    unsigned long sync_rate;       /* master sends sync at this rate */
    unsigned long clock_hz;        /* master clock this frequency (tick rate) */
    unsigned long mask;            /*extra info*/
    unsigned long syncReportThreshold;   /*for sync errors*/
    unsigned long syncUnixReportThreshold; /*for TS_STAMP_FROM_UNIXerrors*/
};
typedef struct TSstampTransStruct TSstampTrans;

/*
 * Interrupt-level communication
 */
static volatile int eventsGood;
static volatile int consecutiveEventReceiverErrorCount;

/*
 * Internal variables
 */
static int             broadcastsGood;
static int             erTimeProviderDebug;
static epicsTimeStamp *eventTable;
static int             eventTableCapacity;
static epicsTimeStamp  timeAtLastBroadcast;
static int             systemTicksAtLastBroadcast;
static int             systemTickRate;
static int              nanosecondsPerSystemTick;

/*
 * Register event receiver methods
 */
static struct drvTSops tsOps, *devTSops;
void drvTsRegisterDeviceSupport(struct drvTSops *ops)
{
    if (devTSops)
        printf("APS event receiver time provider -- duplicate event providers!\n");
    devTSops = ops;
}


/*
 * Return the current time using event-receiver provider.
 */
static int
erGetTime(epicsTimeStamp *pDest)
{
    epicsTimeStamp ts;
    unsigned long ticks;
    int key;

    key = epicsInterruptLock();
    if (!broadcastsGood || !eventsGood) {
        epicsInterruptUnlock(key);
        return epicsTimeERROR;
    }
    ErGetTicks(0, &ticks);
    *pDest = eventTable[ER_RESET_TICKS_EVENT];
    epicsInterruptUnlock(key);
    ts.secPastEpoch = ticks / EVENT_SYSTEM_CLOCK_RATE;
    ts.nsec = (ticks - (ts.secPastEpoch * EVENT_SYSTEM_CLOCK_RATE)) *
                                            NANOSECONDS_PER_EVENT_CLOCK_TICK;
    pDest->secPastEpoch += ts.secPastEpoch;
    pDest->nsec += ts.nsec;
    if(pDest->nsec >= 1000000000) {
        pDest->secPastEpoch++;
        pDest->nsec -= 1000000000;
    }
    return epicsTimeOK;
}

/*
 * Return the time when an event occurred.
 */
static int
erGetEventTime(epicsTimeStamp *pDest, int eventNumber)
{
    int key;

    if ((eventNumber <= 0)
     || (eventNumber >= eventTableCapacity))
        return epicsTimeERROR;
    key = epicsInterruptLock();
    if (!broadcastsGood || !eventsGood) {
        epicsInterruptUnlock(key);
        return epicsTimeERROR;
    }
    *pDest = eventTable[eventNumber];
    epicsInterruptUnlock(key);
    return epicsTimeOK;
}

/*
 * Return the current time using event-broadcast provider.
 */
static int
ebGetTime(epicsTimeStamp *pDest)
{
    epicsTimeStamp base, interval;
    epicsUInt32 lastTicks, currentTicks, intervalTicks;
    int key;

    key = epicsInterruptLock();
    if (!broadcastsGood) {
        epicsInterruptUnlock(key);
        return epicsTimeERROR;
    }
    base = timeAtLastBroadcast;
    currentTicks = osdTickGet();
    lastTicks = systemTicksAtLastBroadcast;
    epicsInterruptUnlock(key);
    /*
     * Note that the following calculation works
     * even when currentTicks has wrapped around
     */
    intervalTicks = currentTicks - lastTicks;
    interval.secPastEpoch = intervalTicks / systemTickRate;
    interval.nsec = (intervalTicks - interval.secPastEpoch * systemTickRate) * nanosecondsPerSystemTick;
    if (interval.secPastEpoch > BROADCAST_TIMEOUT_SECONDS) {
        broadcastsGood = 0;
        return epicsTimeERROR;
    }
    pDest->secPastEpoch = base.secPastEpoch + interval.secPastEpoch;
    pDest->nsec = base.nsec + interval.nsec;
    if(pDest->nsec >= 1000000000) {
        pDest->secPastEpoch++;
        pDest->nsec -= 1000000000;
    }
    return epicsTimeOK;
}

/*	
 * Receive events from event system -- update the event table 
 * Note: called at interrupt level!
 */
static void
eventHandler(int Card, int EventNum, unsigned long ticks)
{
    epicsTimeStamp ts, *st;
    static volatile int lastEventTicks;

    if (eventsGood && (ticks < lastEventTicks)) {
        lastEventTicks = ticks;
        if (broadcastsGood && (--broadcastsGood == 0))
            epicsInterruptContextMessage(
               "***errorHandler: event system error: No packets, or no time change, from master");
    }
    lastEventTicks = ticks;
    eventsGood = 1;
    consecutiveEventReceiverErrorCount = 0;
    
    /* calculate a time stamp from the event tick count */
    ts.secPastEpoch = ticks / EVENT_SYSTEM_CLOCK_RATE;
    ts.nsec = (ticks - (ts.secPastEpoch * EVENT_SYSTEM_CLOCK_RATE)) *
                                            NANOSECONDS_PER_EVENT_CLOCK_TICK;
    /* obtain a copy of the last sync time */
    st = &eventTable[ER_RESET_TICKS_EVENT];
    /* update the event table */
    eventTable[EventNum].secPastEpoch = st->secPastEpoch + ts.secPastEpoch;
    eventTable[EventNum].nsec = st->nsec + ts.nsec;
    /* adjust seconds if needed */
    if(eventTable[EventNum].nsec >= 1000000000) {
        eventTable[EventNum].secPastEpoch++;
        eventTable[EventNum].nsec -= 1000000000;
    }
}

/*
 * Receive error reports from event system
 * Note: called at interrupt level!
 */
static void
errorHandler(int Card, int ErrorNum)
{
    eventsGood = 0;
    if (consecutiveEventReceiverErrorCount < 4) {
        consecutiveEventReceiverErrorCount++;
        switch(ErrorNum)
        {
        case 1:
            epicsInterruptContextMessage(
               "***errorHandler: event system error: TAXI violation");
            break;
        case 2:
            epicsInterruptContextMessage(
               "***errorHandler: event system error: lost heartbeat");
            break;
        case 3:
            epicsInterruptContextMessage(
                "***errorHandler: event system error: lost events");
            break;
        default:
            epicsInterruptContextMessage(
                "***errorHandler: unknown error from event system");
        }
    }
}

/*	
 * Listen for time stamps from master.
 */
static void
timeBroadcastMonitor(void *arg)
{
    TSstampTrans stran;
    struct sockaddr_in sin;
    struct sockaddr fs;
    int i;
    osiSocklen_t fl;
    SOCKET s;
    epicsTimeStamp masterTime, oldMasterTime, myTime;
    int key;
    static int firstUpdate = 1;
    
    oldMasterTime.secPastEpoch = 0;
    oldMasterTime.nsec = 0;
    memset(&sin, 0, sizeof(sin));
    sin.sin_port = htons(TS_SLAVE_PORT);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    if ((s = epicsSocketCreate(AF_INET,SOCK_DGRAM, 0)) < 0) {
        errlogPrintf("erTimeProvider: Can't create socket: %s -- terminating!\n", strerror(SOCKERRNO));
        return;
    }
    if (bind(s, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
        errlogPrintf("erTimeProvider: Can't bind socket: %s -- terminating!\n", strerror(SOCKERRNO));
        close(s);
        return;
    }
    for (;;) {
        fl = sizeof(fs);
        i = recvfrom(s, (char*)&stran, sizeof(stran), 0, &fs, &fl);
        if (i < 0) {
            errlogPrintf("erTimeProvider: Recv failed: %s\n", strerror(SOCKERRNO));
            continue;
        }
        if (i == (sizeof(stran) - 3 * sizeof(unsigned long))) {
            /* R3.13 master doesn't send entire packet */
            stran.mask = htonl(0);
            stran.syncReportThreshold = htonl(0);
            stran.syncUnixReportThreshold = htonl(0);
        }
        else if (i != sizeof(stran)) {
            errlogPrintf("erTimeProvider: received %d, expected %u.\n", i, (unsigned int)sizeof(stran));
            continue;
        }
        if (ntohl(stran.magic) != TS_MAGIC) {
            errlogPrintf("erTimeProvider: invalid packet received.\n");
            continue;
        }
        if (erTimeProviderDebug >= 1000) {
            eventsGood = 0;   /* Simulate faulty event system */
            if (erTimeProviderDebug >= 2000)
                continue;     /* Simulate faulty master */
        }
        masterTime.secPastEpoch = ntohl(stran.master_time.tv_sec);
        masterTime.nsec = ntohl(stran.master_time.tv_nsec);
        if (erTimeProviderDebug >= 2) {
            char cbuf[35];
            epicsTimeToStrftime(cbuf, sizeof cbuf, "%Y/%m/%d %H:%M:%S.%06f", &masterTime);
            printf("erTimeProvider received message from master: %s\n", cbuf);
        }
        if ((masterTime.secPastEpoch == oldMasterTime.secPastEpoch)
         && (masterTime.nsec == oldMasterTime.nsec))
            continue;       /* Master seems to have locked up */
        oldMasterTime = masterTime;
        broadcastsGood = 2;
        key = epicsInterruptLock();
        systemTicksAtLastBroadcast = osdTickGet();
        timeAtLastBroadcast = masterTime;
        myTime = eventTable[ER_RESET_TICKS_EVENT];
        epicsInterruptUnlock(key);
        if (!eventsGood)
            continue;
        if ((myTime.secPastEpoch != masterTime.secPastEpoch)
         || (myTime.nsec != masterTime.nsec)) {
            unsigned long mask;
            unsigned long syncReportThreshold;
            unsigned long syncUnixReportThreshold;
            double        reportThreshold;
            double        unixReportThreshold;
            double timeDiff;
            char cbuf[35];

            timeDiff = epicsTimeDiffInSeconds(&masterTime, &myTime);
            mask = ntohl(stran.mask);
            syncReportThreshold = ntohl(stran.syncReportThreshold);
            syncUnixReportThreshold = ntohl(stran.syncUnixReportThreshold);
            reportThreshold = syncReportThreshold/1000.0;
            unixReportThreshold = syncUnixReportThreshold/1000.0;
            key = epicsInterruptLock();
            eventTable[ER_RESET_TICKS_EVENT].secPastEpoch = masterTime.secPastEpoch;
            eventTable[ER_RESET_TICKS_EVENT].nsec = masterTime.nsec;
            epicsInterruptUnlock(key);
            epicsTimeToStrftime(cbuf, sizeof cbuf, "%Y/%m/%d %H:%M:%S.%06f", &masterTime);
            if (firstUpdate) {
                firstUpdate = 0;
            }
            else if(mask & TS_STAMP_FROM_UNIX) {
                if(fabs(timeDiff) >= unixReportThreshold) {
                   errlogPrintf("erTimeProvider: NTP adjust %f seconds to %s\n",
                                                                timeDiff, cbuf);
                }
            }
            else {
                if(fabs(timeDiff) >= reportThreshold) {
                   errlogPrintf("erTimeProvider: Adjust %f seconds to %s\n",
                                                                timeDiff, cbuf);
                }
            }
        }
    }
}

void ErTimeProviderConfigure(int priority)
{
    static int firstTime = 1;

    if (!firstTime)
        return;
    firstTime = 0;
    systemTickRate = osdTickRateGet();
    nanosecondsPerSystemTick = 1000000000 / systemTickRate;
    if (priority == 0)
        priority = 70;

    if (devTSops != NULL) {
        tsOps = *devTSops;
        if ((tsOps.TShaveReceiver != NULL)
         && ((eventTableCapacity = tsOps.TShaveReceiver(0)) > 0)) {
            /* allocate the event table */
            eventTable = callocMustSucceed(eventTableCapacity, sizeof(*eventTable), "ErTimeProviderConfigure");
            
            if(tsOps.TSregisterEventHandler(0, eventHandler) != 0)
                printf("Failed to register event handler\n");
            else if(tsOps.TSregisterErrorHandler(0, errorHandler) != 0)
                printf("Failed to register error handler\n");
            else {
                generalTimeCurrentTpRegister("APS Event Receiver", priority, erGetTime);
                generalTimeEventTpRegister("APS Event Receiver", priority, erGetEventTime);
            }
        }
    }
    generalTimeCurrentTpRegister("APS Time Broadcast", priority+5, ebGetTime);
    if (epicsThreadCreate("timeBroadcastMonitor",
                       epicsThreadPriorityHigh,
                       epicsThreadGetStackSize(epicsThreadStackSmall),
                       (EPICSTHREADFUNC)timeBroadcastMonitor,
                       NULL) == 0) {
        errlogPrintf("Failed to start time-broadcast monitor client\n");
        return;
    }
}

static const iocshArg ErTimeProviderConfigureArg0 = {"priority", iocshArgInt};
static const iocshArg *const ErTimeProviderConfigureArgs[] = { &ErTimeProviderConfigureArg0 };
static const iocshFuncDef ErTimeProviderConfigureDef = {"ErTimeProviderConfigure",1,ErTimeProviderConfigureArgs};
static void ErTimeProviderConfigureCall(const iocshArgBuf * args)
{
    ErTimeProviderConfigure(args[0].ival);
}

/*
 * Allow applications to continue to invoke TSconfigure() in st.cmd scripts.
 */
static const iocshArg TSconfigureArg0 = {"master", iocshArgInt};
static const iocshArg TSconfigureArg1 = {"sync_rate_sec", iocshArgInt};
static const iocshArg TSconfigureArg2 = {"clock_rate_hz", iocshArgInt};
static const iocshArg TSconfigureArg3 = {"master_port", iocshArgInt};
static const iocshArg TSconfigureArg4 = {"slave_port", iocshArgInt};
static const iocshArg TSconfigureArg5 =
    {"millisecond_request_time_out", iocshArgInt};
static const iocshArg TSconfigureArg6 = {"type", iocshArgInt};
static const iocshArg *const TSconfigureArgs[] = {
    &TSconfigureArg0,&TSconfigureArg1,&TSconfigureArg2,
    &TSconfigureArg3,&TSconfigureArg4,&TSconfigureArg5,
    &TSconfigureArg6};
static const iocshFuncDef TSconfigureDef = {"TSconfigure",7,TSconfigureArgs};
static void TSconfigureCall(const iocshArgBuf * args) {
    printf("*************************************\n"
           "* TSconfigure() is deprecated       *\n"
           "* Change your script to use         *\n"
           "* ErTimeProviderConfigure() instead *\n"
           "*************************************\n");
           
    if (args[6].ival != 0)
        devTSops = 0;
    ErTimeProviderConfigure(0);
}

static void
erTimeProviderRegister(void)
{
    static int firstTime = 1;
    if(firstTime) {
        firstTime = 0;
        iocshRegister(&ErTimeProviderConfigureDef,ErTimeProviderConfigureCall);
        iocshRegister(&TSconfigureDef,TSconfigureCall);
    }
}
epicsExportAddress(int,erTimeProviderDebug);
epicsExportRegistrar(erTimeProviderRegister);
